# Copyright 2018 ML2Grow BVBA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from umongo.instance import LazyLoaderInstance
from .builder import DataStoreBuilder
from .helpers import DataStoreClientWrapper


class DataStoreInstance(LazyLoaderInstance):

    def __init__(self, *args, **kwargs):
        self.BUILDER_CLS = DataStoreBuilder  # pylint: disable=C0103
        super().__init__(*args, **kwargs)

    def init(self, db):
        super(DataStoreInstance, self).init(DataStoreClientWrapper(db))



from umongo.data_proxy import BaseDataProxy, BaseNonStrictDataProxy
from google.cloud import datastore


class KeyTranslatingBaseDataProxy(BaseDataProxy):
    """
    Handles converting fields stored as datastore.Key (because attribute='_id' was set) back to their value
    Only for retrieving data, wrapping the keys here is difficult (the other key attributes are unknown in
    the DataProxy)
    """
    # pylint: disable=W0201
    def dump(self):
        data_old = self._data
        self._data = {}
        for key, value in data_old.items():
            if isinstance(value, datastore.Key):
                value = value.id_or_name
            self._data[key] = value
        result = super().dump()
        self._data = data_old
        return result

    def get(self, name, to_raise=KeyError):
        value = super().get(name, to_raise)
        if isinstance(value, datastore.Key):
            value = value.id_or_name
        return value


class KeyTranslatingBaseNonStrictDataProxy(BaseNonStrictDataProxy):
    # pylint: disable=W0201
    def dump(self):
        data_old = self._data
        self._data = {}
        for key, value in data_old.items():
            if isinstance(value, datastore.Key):
                value = value.id_or_name
            self._data[key] = value
        result = super().dump()
        self._data = data_old
        return result

    def get(self, name, to_raise=KeyError):
        value = super().get(name, to_raise)
        if isinstance(value, datastore.Key):
            value = value.id_or_name
        return value


def data_proxy_factory(basename, schema, strict=True):
    cls_name = "%sDataProxy" % basename

    nmspc = {
        '__slots__': (),
        'schema': schema,
        '_fields': schema.fields,
        '_fields_from_mongo_key': {v.attribute or k: v for k, v in schema.fields.items()}
    }

    data_proxy_cls = type(
        cls_name,
        (KeyTranslatingBaseDataProxy if strict else KeyTranslatingBaseNonStrictDataProxy, ),
        nmspc
    )
    return data_proxy_cls



# Copyright 2018 ML2Grow BVBA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

from umongo import fields
from umongo.frameworks.pymongo import _list_io_validate, _embedded_document_io_validate
from umongo.builder import _build_document_opts as _build_document_opts_orig
from umongo.fields import ListField, EmbeddedField
from umongo.document import DocumentImplementation, DocumentOpts
from umongo.builder import (
    BaseBuilder,
    Schema,
    DocumentTemplate,
    _collect_indexes,
    _collect_schema_attrs,
    on_need_add_id_field,
    add_child_field
)

from .data_proxy import data_proxy_factory
from .helpers import DataStoreClientWrapper
from .document import DataStoreDocument
from .reference import DataStoreReference
from .fields import ReferenceField, SUPPORTED_FIELD_TYPES


def _build_document_opts(instance, template, name, nmspc, bases):
    opts = _build_document_opts_orig(instance, template, name, nmspc, bases)
    components = opts.__dict__
    # Override camel_to_snake function
    components['collection_name'] = name
    return DocumentOpts(**components)


class DataStoreBuilder(BaseBuilder):

    BASE_DOCUMENT_CLS = DataStoreDocument

    @staticmethod
    def is_compatible_with(database):
        return isinstance(database, DataStoreClientWrapper)

    def _patch_field(self, field):
        super()._patch_field(field)

        validators = field.io_validate
        if not validators:
            field.io_validate = []
        else:
            if hasattr(validators, '__iter__'):
                validators = list(validators)
            else:
                validators = [validators]
            field.io_validate = validators
        if isinstance(field, ListField):
            field.io_validate_recursive = _list_io_validate
        if isinstance(field, fields.ReferenceField):
            # due to eventual consistency, this check is to prone to failure.
            #field.io_validate.append(_reference_io_validate)
            field.reference_cls = DataStoreReference
        if isinstance(field, EmbeddedField):
            field.io_validate_recursive = _embedded_document_io_validate

    @classmethod
    def _convert_reference_field(cls, field):
        # Swap referencefield with our implementation for datastore keys
        if isinstance(field, fields.ReferenceField):
            field = ReferenceField(**field.__dict__)
        if isinstance(field, ListField):
            field.container = cls._convert_reference_field(field.container)
        return field

    @classmethod
    def _check_field(cls, field):
        if field.__class__ not in SUPPORTED_FIELD_TYPES:
            raise Exception("Field type {0} is currently unsupported for datastore".format(field.__class__.__name__))
        if isinstance(field, ListField):
            field.container = cls._check_field(field.container)
        return field

    @classmethod
    def _apply_to_schema(cls, schema_fields, func):
        transformed_fields = {}
        for name, field in schema_fields.items():
            transformed_fields[name] = func(field)
        return transformed_fields

    def build_document_from_template(self, template):
        """
        Lots of copy paste, only to sneak in _apply_to_schema which converts ReferenceFields to our
        ReferenceFields, and checks valid fields are used.
        """
        assert issubclass(template, DocumentTemplate)
        name = template.__name__
        bases = self._convert_bases(template.__bases__)
        opts = _build_document_opts(self.instance, template, name, template.__dict__, bases)
        nmspc, schema_fields, schema_non_fields = _collect_schema_attrs(template.__dict__)

        # Things we add here:
        # Check if only supported fields are used
        schema_fields = self._apply_to_schema(schema_fields, self._check_field)
        # Convert umongo.fields.ReferenceField to our implementation
        schema_fields = self._apply_to_schema(schema_fields, self._convert_reference_field)
        nmspc['opts'] = opts

        # Create schema by retrieving inherited schema classes
        schema_bases = tuple([base.Schema for base in bases
                              if hasattr(base, 'Schema')])
        if not schema_bases:
            schema_bases = (Schema, )
        on_need_add_id_field(schema_bases, schema_fields)
        # If Document is a child, _cls field must be added to the schema
        if opts.is_child:
            add_child_field(name, schema_fields)
        schema_cls = self._build_schema(template, schema_bases, schema_fields, schema_non_fields)
        nmspc['Schema'] = schema_cls
        schema = schema_cls()
        nmspc['schema'] = schema
        nmspc['DataProxy'] = data_proxy_factory(name, schema, strict=opts.strict)

        # _build_document_opts cannot determine the indexes given we need to
        # visit the document's fields which weren't defined at this time
        opts.indexes = _collect_indexes(nmspc.get('Meta'), schema.fields, bases)

        implementation = type(name, bases, nmspc)
        self._templates_lookup[template] = implementation
        # Notify the parent & grand parents of the newborn !
        for base in bases:
            for parent in inspect.getmro(base):
                if (not issubclass(parent, DocumentImplementation) or
                        parent is DocumentImplementation):
                    continue
                parent.opts.offspring.add(implementation)
        return implementation



# Copyright 2018 ML2Grow BVBA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from umongo.document import DocumentImplementation
from umongo.exceptions import NotCreatedError, ValidationError, DeleteError
from umongo.frameworks.pymongo import _io_validate_data_proxy

from .helpers import cook_find_filter


class DataStoreDocument(DocumentImplementation):

    """ The actual framework implementation class. """

    __slots__ = ()

    opts = DocumentImplementation.opts

    def reload(self):
        """
        Retrieve and replace document's data by the ones in database.
        """
        if not self.is_created:
            raise NotCreatedError("Document doesn't exists in database")
        ret = self.get(self.pk.id_or_name)
        if ret is None:
            raise NotCreatedError("Document doesn't exists in database")
        self._data = self.DataProxy()
        self._data.from_mongo(ret)

    def commit(self, io_validate_all=False):
        """
        Commit the document in database.
        If the document doesn't already exist it will be inserted, otherwise
        it will be updated.
       """
        return self.commit_multi([self], io_validate_all=io_validate_all)

    @classmethod
    def commit_multi(cls, docs, io_validate_all=False):
        # pylint: disable=W0212
        payloads = []
        try:
            for doc in docs:
                if doc.is_modified():
                    doc.required_validate()
                    doc.io_validate(validate_all=io_validate_all)
                    payloads.append(doc._data.to_mongo(update=False))

            keys = cls.collection.put_multi(payloads)  # pylint: disable=E1101

            for key, doc in zip(keys, docs):
                if not doc.is_created:
                    doc._data.set_by_mongo_name('_id', key)

        except Exception as exc:
            # Need to dig into error message to find faulting index
            raise ValidationError(str(exc))

        for doc in docs:
            doc.is_created = True
            doc._data.clear_modified()

    def delete(self):
        """
        Remove the document from database.
        """
        if not self.is_created:
            raise NotCreatedError("Document doesn't exists in database")
        try:
            self.collection.delete(self.pk)  # pylint: disable=E1101
            self.is_created = False
        except Exception as exc:
            raise DeleteError(str(exc))

    def io_validate(self, validate_all=False):
        """
        Run the io_validators of the document's fields.
        """
        if validate_all:
            return _io_validate_data_proxy(self.schema, self._data)
        else:
            return _io_validate_data_proxy(
                self.schema, self._data, partial=self._data.get_modified_fields())

    @classmethod
    def find_one(cls, filters=None, order=()):
        """
        Find a single document in database.
        """
        try:
            doc = next(cls.find(filters, limit=1, order=order))
        except StopIteration:
            doc = None
        return doc

    @classmethod
    def find(cls, filters=None, order=(), limit=None):
        """
        Find a list document in database.

        Returns a cursor that provide Documents.
        """
        filters = cook_find_filter(cls, filters or {})

        for ret in cls.collection.query(filters, order=order, limit=limit):  # pylint: disable=E1101
            yield cls.build_from_mongo(ret, use_cls=True)

    @classmethod
    def count(cls, filters=None):
        """
        Get the number of documents in this collection.
        """
        filters = cook_find_filter(cls, filters or {})
        return len(list(cls.find(filters)))

    @classmethod
    def ensure_indexes(cls):
        """
        No index support for datastore
        """
        return
        yield  # pylint: disable=W0101

    @classmethod
    def get(cls, key):
        return cls.get_multi([key])[0]

    @classmethod
    def get_multi(cls, pks):
        returned = cls.collection.get_multi(pks)  # pylint: disable=E1101
        return list(map(lambda r: cls.build_from_mongo(r, use_cls=True), returned))



# Copyright 2018 ML2Grow BVBA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from umongo.data_objects import Reference
from umongo.exceptions import ValidationError


class DataStoreReference(Reference):
    """ Reference representation """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._document = None

    def fetch(self, no_data=False, force_reload=False):
        if not self._document or force_reload:
            if self.pk is None:
                raise ReferenceError('Cannot retrieve a None Reference')
            self._document = self.document_cls.find_one({'_id': self.pk})
            if not self._document:
                raise ValidationError(self.error_messages['not_found'].format(document=self.document_cls.__name__))
        return self._document



# Copyright 2018 ML2Grow BVBA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import itertools
from datetime import datetime

from google.cloud import datastore
from umongo.frameworks import tools


def cook_find_filter(doc_cls, filters):
    filters = tools.cook_find_filter(doc_cls, filters)
    key = filters.pop('_id', None)
    if key:
        filters['__key__'] = key
    return filters


def _apply_filter(query, *args, fresh=False):
    if fresh:
        query = datastore.Query(
            client=query._client,  # pylint: disable=W0212
            kind=query.kind,
            project=query.project,
            namespace=query.namespace,
            ancestor=query.ancestor,
            filters=copy.deepcopy(query.filters),
            projection=query.projection,
            order=query.order,
            distinct_on=query.distinct_on
        )
    query.add_filter(*args)
    return query


class DataStoreClientWrapper:
    """ For compatibility with the collection property in umongo """
    def __init__(self, client):
        super().__init__()
        self.client = client

    def __getitem__(self, cname) -> 'CollectionAbstraction':
        return CollectionAbstraction(self.client, cname)


class CollectionAbstraction:
    """
    Represents a "collection", as a convention we interpret this as an Entity kind.

    Instances are returned by the collection property, has some methods to interact
    with datastore on this specific entity.
    """

    _filter_oper_lookup = {
        '$gt': lambda q, k, v: [_apply_filter(q, k, '>', v)],
        '$lt': lambda q, k, v: [_apply_filter(q, k, '<', v)],
        '$eq': lambda q, k, v: [_apply_filter(q, k, '=', v)],
        '$gte': lambda q, k, v: [_apply_filter(q, k, '>=', v)],
        '$lte': lambda q, k, v: [_apply_filter(q, k, '<=', v)],
        '$in': lambda q, k, v: [_apply_filter(q, k, '=', i, fresh=True) for i in v]
    }

    def __init__(self, client, cname):
        super(CollectionAbstraction, self).__init__()
        self.client = client
        self.cname = cname

    @staticmethod
    def _unpack(entity):
        if entity is None:
            return None

        payload = {
            '_id': entity.key
        }

        for key, value in entity.items():
            if isinstance(value, datetime):
                value = value.replace(tzinfo=None)
            payload[key] = value
        return payload

    def _pack(self, payload):
        ref = payload.pop('_id', None)
        if not isinstance(ref, datastore.Key):
            ref = self.key(ref)
        entity = datastore.Entity(key=ref)
        entity.update(payload)
        return entity

    def key(self, ref=None):
        if ref:
            return self.client.key(self.cname, ref)
        else:
            return self.client.key(self.cname)

    def get(self, key):
        return self.get_multi([key])[0]

    def get_multi(self, keys):
        keys_wrapped = list(map(self.key, keys))
        entities = self.client.get_multi(keys_wrapped)
        entity_map = {e.key.id_or_name: e for e in entities}
        return [self._unpack(entity_map.get(key, None)) for key in keys]

    def put(self, payload):
        keys = self.put_multi([payload])
        return keys[0]

    def put_multi(self, payloads):
        entities = list(map(self._pack, payloads))
        self.client.put_multi(entities)
        return list(map(lambda k: k.key, entities))

    def delete(self, key):
        self.client.delete(key)

    def query(self, filters, limit=None, order=()):
        queries = [self.client.query(kind=self.cname, order=order)]
        for attr, domain in filters.items():
            if isinstance(domain, dict):
                for oper, operand in domain.items():
                    query_groups = [self._filter_oper_lookup[oper](query, attr, operand) for query in queries]
                    queries = itertools.chain(*query_groups)
            else:
                for query in queries:
                    query.add_filter(attr, '=', domain)

        for query in queries:
            for ret in query.fetch(limit=limit):
                yield self._unpack(ret)



# Copyright 2018 ML2Grow BVBA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from setuptools import setup, find_packages


def read_version(fname):
    verstrline = open(fname, "rt").read()
    vsre = r"^__version__ = ['\"]([^'\"]*)['\"]"
    matched = re.search(vsre, verstrline, re.M)
    if matched:
        return matched.group(1)
    else:
        raise RuntimeError("Unable to find version string in %s." % (fname,))


REQUIREMENTS = [
    'umongo',
    'pymongo',
    'google-cloud-datastore>=1.7.0'
]

with open('README.md') as f:
    README = f.read()

setup(
    name='udatastore',
    version=read_version("udatastore/_version.py"),
    url='https://github.com/ml2grow/udatastore',
    license="Apache License 2.0",
    author='javdrher',
    author_email='joachim@ml2grow.com',
    description='Datastore framework implementation for umongo',
    long_description=README,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=('tests*',)),
    extras_require={
        'test': ['nox']
    },
    include_package_data=True,
    install_requires=REQUIREMENTS,
    test_suite='tests',
    keywords='udatastore umongo pymongo datastore',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ]
)



# Copyright 2018 ML2Grow BVBA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle

import umongo
from umongo.exceptions import ValidationError
from umongo.data_objects import Reference
from marshmallow import fields as ma_fields
from marshmallow import missing


class ReferenceField(umongo.fields.ReferenceField):
    """
    umongo.fields.ReferenceField inherits the serialize/deserialize behaviour which creates bson.ObjectIDs

    In order to avoid this, we must override the _(de)serialize of the ReferenceField class.
    This involves some copy paste.
    """

    def _deserialize(self, value, attr, data):
        if value is None:
            return None
        elif isinstance(value, Reference):
            if value.document_cls != self.document_cls:
                raise ValidationError("`{document}` reference expected.".format(document=self.document_cls.__name__))
            if not isinstance(value, self.reference_cls):
                value = self.reference_cls(value.document_cls, value.pk)
            return value
        elif isinstance(value, self.document_cls):
            if not value.is_created:
                raise ValidationError("Cannot reference a document that has not been created yet.")
            value = value.pk
        elif isinstance(value, self._document_implementation_cls):
            raise ValidationError("`{document}` reference expected.".format(document=self.document_cls.__name__))
        # `value` is similar to data received from the database so we
        # can use `_deserialize_from_mongo` to finish the deserialization
        return self._deserialize_from_mongo(value)


class _MaBytesField(ma_fields.Field):
    def _serialize(self, value, attr, obj):
        if value is None:
            return None
        return pickle.dumps(value)

    def _deserialize(self, value, attr, data):
        if isinstance(value, bytes):
            return pickle.loads(value)
        return value


class BytesField(umongo.abstract.BaseField, _MaBytesField):
    def _serialize_to_mongo(self, obj):
        if obj is None:
            return missing
        return pickle.dumps(obj)

    def _deserialize_from_mongo(self, value):
        if value is None:
            return None
        return pickle.loads(value)


SUPPORTED_FIELD_TYPES = [
    umongo.fields.BooleanField,
    umongo.fields.DateTimeField,
    umongo.fields.ReferenceField,
    umongo.fields.StringField,
    umongo.fields.NumberField,
    umongo.fields.IntegerField,
    umongo.fields.UrlField,
    umongo.fields.EmailField,
    umongo.fields.EmbeddedField,
    umongo.fields.ListField,
    umongo.fields.DictField,
    umongo.fields.FormattedStringField,
    umongo.fields.FloatField,
    ReferenceField,
    BytesField
]
